import time
import traceback

import schedule

from lifeguard.logger import lifeguard_logger as logger
from lifeguard.validations import VALIDATIONS

VALID_TIME_PERIODS = [
    "seconds",
    "minutes",
    "hours",
    "days",
    "weeks",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]

MOMENTS = [
    "second",
    "minute",
    "hour",
    "day",
    "week",
    "monday",
    "tuesday",
    "wednesday",
    "thursday",
    "friday",
    "saturday",
    "sunday",
]


def configure_validations():
    for validation in VALIDATIONS:
        content = VALIDATIONS[validation]
        if "every" in content["schedule"]:
            time_period = get_time_period(content)
            if time_period in VALID_TIME_PERIODS:
                job_instance = schedule.every(content["schedule"]["every"][time_period])
                job_func = get_dynamic_job_func(job_instance, time_period)
                job_func.do(content["ref"])
        if "at" in content["schedule"]:
            time_moment = get_time_moment(content)
            if time_moment in MOMENTS:
                moment = getattr(schedule.every(), time_moment)
                moment.at(content["schedule"]["at"][time_moment]).do(content["ref"])


def get_dynamic_job_func(job_instance, time_period):
    return getattr(job_instance, time_period)


def get_time_period(content):
    keys = content["schedule"]["every"].keys()
    time_period = list(keys)[0]
    return time_period


def get_time_moment(content):
    keys = content["schedule"]["at"].keys()
    time_moment = list(keys)[0]
    return time_moment


def start_scheduler():
    configure_validations()
    while True:
        time.sleep(1)
        try:
            schedule.run_pending()
        except Exception as exception:
            logger.error(
                "error on execute scheduler %s",
                str(exception),
                extra={"traceback": traceback.format_exc()},
            )



"""
Common assets controllers
"""
import io
from os.path import join

from lifeguard.controllers import Response, register_custom_controller, send_status
from lifeguard.logger import lifeguard_logger as logger
from lifeguard.settings import LIFEGUARD_DIRECTORY


def send_file(path, content_type):
    with open(join(LIFEGUARD_DIRECTORY, "public", path), "rb") as file:
        response = Response()
        response.content = io.BytesIO(file.read())
        response.content_type = content_type
        return response


def send_css(path):
    try:
        logger.info("returning css %s", path)
        return send_file("css/{}".format(path), "text/css")
    except:
        return send_status(404)


def send_js(path):
    try:
        logger.info("returning js %s", path)
        return send_file("js/{}".format(path), "application/javascript")
    except:
        return send_status(404)


def send_img(path):
    try:
        logger.info("returning img %s", path)
        return send_file("img/{}".format(path), "image/jpeg")
    except:
        return send_status(404)


def load_assets_controllers():
    register_custom_controller("/img/<path:path>", send_img, {"methods": ["GET"]})
    register_custom_controller("/js/<path:path>", send_js, {"methods": ["GET"]})
    register_custom_controller("/css/<path:path>", send_css, {"methods": ["GET"]})



"""
Lifeguard validation core
"""
import importlib
import sys

from lifeguard.logger import lifeguard_logger as logger
from lifeguard.statuses import NORMAL, WARNING, PROBLEM, ACTION_STATUSES, change_status
from lifeguard.settings import (
    LIFEGUARD_DIRECTORY,
    SETTINGS_MANAGER,
    LIFEGUARD_CONTROLLERS_ASSETS_ENABLED,
)
from lifeguard.controllers import load_custom_controllers
from lifeguard.controllers.assets import load_assets_controllers
from lifeguard.validations import load_validations


def recover_settings():
    """
    Get settings.py from root of project
    """

    logger.info("lifeguard directory %s", LIFEGUARD_DIRECTORY)
    sys.path.append(LIFEGUARD_DIRECTORY)

    logger.info("path: %s", sys.path)

    lifeguard_settings = "lifeguard_settings"
    logger.info("loading %s", lifeguard_settings)
    if importlib.util.find_spec(lifeguard_settings) is None:
        logger.error("lifeguard_settings.py not found!!!")
        sys.exit(-1)
    return __import__(lifeguard_settings)


def setup(lifeguard_context):
    """
    Setup lifeguard context
    """
    # init plugins
    lifeguard_settings = recover_settings()

    for plugin in lifeguard_settings.PLUGINS:
        SETTINGS_MANAGER.settings.update(plugin.settings.SETTINGS_MANAGER.settings)
        if not lifeguard_context.only_settings:
            plugin.init(lifeguard_context)

    load_custom_controllers()

    if LIFEGUARD_CONTROLLERS_ASSETS_ENABLED:
        load_assets_controllers()

    load_validations()

    lifeguard_settings.setup(lifeguard_context)



import unittest

from unittest.mock import patch

from lifeguard import NORMAL
from lifeguard.validations import (
    ValidationResponse,
    load_validations,
    VALIDATIONS,
)

from tests.fixtures.validations.simple_validation_with_action_on_errors import (
    ON_ERROR_ACTION_MOCK,
)


class TestValidationResponse(unittest.TestCase):
    def test_validation_response_object(self):
        details = {}
        settings = {}

        response = ValidationResponse(NORMAL, details, settings, validation_name="name")

        self.assertEqual(response.validation_name, "name")
        self.assertEqual(response.status, "NORMAL")
        self.assertEqual(response.details, details)
        self.assertEqual(response.settings, settings)

    def test_validation_response_to_str(self):
        response = ValidationResponse(NORMAL, {}, {}, validation_name="name")
        self.assertEqual(
            str(response),
            "{'validation_name': 'name', 'status': 'NORMAL', 'details': {}, 'settings': {}}",
        )

    def test_validate_status_value(self):
        with self.assertRaises(ValueError) as error:
            ValidationResponse("INVALID", {}, {})

        self.assertEquals(str(error.exception), "INVALID is not a valid status")


class TestValidations(unittest.TestCase):
    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_load_validations(self, mock_logger):
        load_validations()
        mock_logger.info.assert_any_call("loading validation %s", "simple_validation")
        self.assertTrue("simple_validation" in VALIDATIONS)

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_load_validations_subdir(self, mock_logger):
        load_validations()
        mock_logger.info.assert_any_call(
            "loading validation %s", "simple_subdir_validation"
        )
        self.assertTrue("simple_subdir_validation" in VALIDATIONS)

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_execute_validation(self, _mock_logger):
        load_validations()
        response = VALIDATIONS["simple_validation"]["ref"]()
        self.assertEqual(response.status, NORMAL)
        self.assertEqual(VALIDATIONS["simple_validation"]["settings"], {})

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    def test_execute_validation_with_action(self, mock_logger):
        load_validations()
        VALIDATIONS["simple_with_action_validation"]["ref"]()
        mock_logger.info.assert_called_with(
            "executing action %s with result %s...",
            "simple_action",
            "{'validation_name': 'simple_with_action_validation', 'status': 'NORMAL', 'details': {}, 'settings': None}",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch(
        "lifeguard.validations.LIFEGUARD_RUN_ONLY_VALIDATIONS",
        ["simple_with_action_validation"],
    )
    @patch("lifeguard.validations.logger")
    def test_execute_validation_because_in_list(self, mock_logger):
        load_validations()
        self.assertIsNotNone(VALIDATIONS["simple_with_action_validation"]["ref"]())
        mock_logger.info.assert_called_with(
            "executing action %s with result %s...",
            "simple_action",
            "{'validation_name': 'simple_with_action_validation', 'status': 'NORMAL', 'details': {}, 'settings': None}",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch(
        "lifeguard.validations.LIFEGUARD_RUN_ONLY_VALIDATIONS",
        ["simple_validation"],
    )
    @patch("lifeguard.validations.logger")
    def test_not_execute_validation_because_not_in_list(self, mock_logger):
        load_validations()
        self.assertIsNone(VALIDATIONS["simple_with_action_validation"]["ref"]())
        mock_logger.info.assert_called_with(
            "validation %s not in LIFEGUARD_RUN_ONLY_VALIDATIONS",
            "simple_with_action_validation",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch(
        "lifeguard.validations.LIFEGUARD_SKIP_VALIDATIONS",
        ["simple_validation"],
    )
    @patch("lifeguard.validations.logger")
    def test_execute_validation_because_not_in_skip_list(self, mock_logger):
        load_validations()
        self.assertIsNotNone(VALIDATIONS["simple_with_action_validation"]["ref"]())
        mock_logger.info.assert_called_with(
            "executing action %s with result %s...",
            "simple_action",
            "{'validation_name': 'simple_with_action_validation', 'status': 'NORMAL', 'details': {}, 'settings': None}",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch(
        "lifeguard.validations.LIFEGUARD_SKIP_VALIDATIONS",
        ["simple_with_action_validation"],
    )
    @patch("lifeguard.validations.logger")
    def test_not_execute_validation_because_in_skip_list(self, mock_logger):
        load_validations()
        self.assertIsNone(VALIDATIONS["simple_with_action_validation"]["ref"]())
        mock_logger.info.assert_called_with(
            "validation %s in LIFEGUARD_SKIP_VALIDATIONS",
            "simple_with_action_validation",
        )

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    @patch("lifeguard.validations.traceback")
    def test_execute_validation_with_invalid_action(self, mock_traceback, mock_logger):
        mock_traceback.format_exc.return_value = "traceback"
        load_validations()
        result = VALIDATIONS["simple_with_invalid_action_validation"]["ref"]()
        mock_logger.error.assert_called_with(
            "validation error %s: %s",
            "simple_with_invalid_action_validation",
            "invalid_action() takes 0 positional arguments but 2 were given",
            extra={"traceback": "traceback"},
        )
        self.assertIsNone(result)

    @patch("lifeguard.validations.LIFEGUARD_DIRECTORY", "tests/fixtures")
    @patch("lifeguard.validations.logger")
    @patch("lifeguard.validations.traceback")
    def test_execute_validation_with_error_and_error_actions(
        self, mock_traceback, mock_logger
    ):
        mock_traceback.format_exc.return_value = "traceback"
        load_validations()
        VALIDATIONS["simple_validation_with_action_on_errors"]["ref"]()

        response = ON_ERROR_ACTION_MOCK.mock_calls[0][1][0]
        settings = ON_ERROR_ACTION_MOCK.mock_calls[0][1][1]

        self.assertEqual(
            response.validation_name, "simple_validation_with_action_on_errors"
        )
        self.assertEqual(
            response.details,
            {
                "exception": "error",
                "traceback": "traceback",
                "use_error_template": True,
            },
        )
        self.assertEqual(response.status, "PROBLEM")
        self.assertEqual(settings, {})



"""Genral Utility Functions
"""
import os


def build_import(path, cur_name):
    """Returns python like import name"""
    head, tail = os.path.split(path)
    if tail:
        cur_name = f"{tail}.{cur_name}"
        final_name = cur_name
    if head:
        final_name = build_import(head, cur_name)
    return final_name
